package metrics

import (
	"net/http"
	"sync"

	"git.frostfs.info/TrueCloudLab/frostfs-s3-gw/internal/logs"
	dto "github.com/prometheus/client_model/go"
	"go.uber.org/zap"
)

type AppMetrics struct {
	logger  *zap.Logger
	gate    *GateMetrics
	mu      sync.RWMutex
	enabled bool
}

func NewAppMetrics(logger *zap.Logger, poolStatistics StatisticScraper, enabled bool) *AppMetrics {
	if !enabled {
		logger.Warn(logs.MetricsAreDisabled)
	}
	return &AppMetrics{
		logger:  logger,
		gate:    NewGateMetrics(poolStatistics),
		enabled: enabled,
	}
}

func (m *AppMetrics) SetEnabled(enabled bool) {
	if !enabled {
		m.logger.Warn(logs.MetricsAreDisabled)
	}

	m.mu.Lock()
	m.enabled = enabled
	m.mu.Unlock()
}

func (m *AppMetrics) State() *StateMetrics {
	if !m.isEnabled() {
		return nil
	}

	return m.gate.State
}

func (m *AppMetrics) Shutdown() {
	m.mu.Lock()
	if m.enabled {
		m.gate.State.SetHealth(HealthStatusShuttingDown)
		m.enabled = false
	}
	m.gate.Unregister()
	m.mu.Unlock()
}

func (m *AppMetrics) isEnabled() bool {
	m.mu.RLock()
	defer m.mu.RUnlock()
	return m.enabled
}

func (m *AppMetrics) Handler() http.Handler {
	return m.gate.Handler()
}

func (m *AppMetrics) Update(user, bucket, cnrID string, reqType RequestType, in, out uint64) {
	if !m.isEnabled() {
		return
	}

	m.gate.Billing.apiStat.Update(user, bucket, cnrID, reqType, in, out)
}

func (m *AppMetrics) Statistic() *APIStatMetrics {
	if !m.isEnabled() {
		return nil
	}

	return m.gate.Stats
}

func (m *AppMetrics) Gather() ([]*dto.MetricFamily, error) {
	return m.gate.Gather()
}

func (m *AppMetrics) MarkHealthy(endpoint string) {
	if !m.isEnabled() {
		return
	}

	m.gate.HTTPServer.MarkHealthy(endpoint)
}

func (m *AppMetrics) MarkUnhealthy(endpoint string) {
	if !m.isEnabled() {
		return
	}

	m.gate.HTTPServer.MarkUnhealthy(endpoint)
}
